@@ -1,13 +1,12 @@
 import numpy as np
-# from multiprocessing import Pool#
 from pathos.multiprocessing import ProcessingPool as Pool
+# from multiprocessing_on_dill import Pool
 import os
 import pickle
 import seaborn as sns
 from .Model import Model
 
 import matplotlib
-matplotlib.use('Agg')
 import matplotlib.pyplot as plt
 
 
@@ -20,6 +19,7 @@ class Swarm(object):
                  Phip = 0.2,         # PSO rule parameter
                  Phig = 0.2,        # PSO rule parameter
                  Mh_fraction= 0.0,
+                 Jitter_weight = 0.0,
                  Tol = 1.0e-3,
                  Maxiter = 1.0e6,
                  Periodic = None,
@@ -33,7 +33,14 @@ class Swarm(object):
                  Saveevolution = False,
                  Velocity_min = None,
                  Velocity_minimum_factor = 100,
-                 Proposalcov = None):
+                 Proposalcov = None,
+                 Initial_guess_v_factor = 3,
+                 Clip_lower_velocity = True,
+                 Clip_upper_velocity = True,
+                 Delta_max = 0.3,
+                 Delta_min = 0.0001,
+                 Velocity_clipping_or_rescale= 'Clip',
+                 Reinitialise_velocities_from_initial_guess=True):
         """
 
         Minimum working example of Particle swarm optimization class.
@@ -58,8 +65,11 @@ class Swarm(object):
             the phi_p parameter, cognitive coefficient for velocity updating [defaults to .2]
         Phig: float
             the phi_g parameter, social coefficient for velocity updating [defaults to .2]
-        Mh_fraction: float:
+        Mh_fraction: float
             parameter controlling proportion of velocity rule dictated by MCMC [defaults to 0.]
+        Jitter_weight: float
+            parameter that controls the movement of velocities to point towards a random particle in the swarm [defaults to 0.]
+             (used to get particle out of local maxima they are stuck in)
         Tol: float
             the tol on the function value between, this is the spread in function values below which
             we consider the optimization algoritm to have converged [defaults to 1.0e-3]
@@ -97,7 +107,23 @@ class Swarm(object):
             Set absolute minimum speed to 1/velocity_minimum_factor of the prior range specified in each parameter,
              Only relevant of velocity min array not provided by user [defaults to 100]
         Proposalcov: numpy array
-            Covariance matrix for gaussian proposal distribution for MH part of velocity rule [defaults to identity]"""
+            Covariance matrix for gaussian proposal distribution for MH part of velocity rule [defaults to identity]
+        Initial_guess_v_factor: float/int
+            Fiddle factor used to multiply initial guess spreads to initialise velocities [defaults to 3]
+        Clip_lower_velocity: boolean
+            Flag for if velocity should be clipped (rescaled) on the lower end preserving direction of travel [defaults to True]
+        Clip_upper_velocity: boolean
+            Flag for if velocity should be clipped (rescaled) on the upper end preserving direction of travel [defaults to True]
+        Delta_max: float
+            fiddle parameter used to control maximum magnitude of velocity vector [defaults to 0.3] [Only used if velocity rescaling enabled]
+        Delta_min: float
+            fiddle parameter used to control minimum mangitude of velocity vector [defaults to 0.0001] [Only used if velocity rescaling enabled]
+        Velocity_clipping_or_rescale: str, either 'Clip' or 'Rescale'
+            flag used to choose prescription for clamping velocities, either clip velocities to some minimum velocity in each dimension, or
+             rescale to some minimum and or maximum velocity magnitudes.
+        Reinitialise_velocities_from_initial_guess: boolean
+            Wether to reinitialise velocities from initial guess (or if false, initialise from v_min array provided by user) [defaults to True]
+        """
 
         self.Ndim = len(Model.names)
 
@@ -118,10 +144,14 @@ class Swarm(object):
 
         self.Tol = Tol
 
+        self.initial_guess_v_factor = Initial_guess_v_factor
+
         self.Omega = Omega
         self.PhiP = Phip
         self.PhiG = Phig
         self.MH_fraction = Mh_fraction
+        self.Jitter_weight  = Jitter_weight
+
 
         if Proposalcov is None:
             self.ProposalCov = np.identity(self.Ndim)
@@ -162,11 +192,28 @@ class Swarm(object):
         self.BoundsArray = np.array(self.Model.bounds)
 
         #     NOTE: ptp by default positive
-        self.velocity_min = Velocity_min
-        self.velocity_minimum_factor = Velocity_minimum_factor
 
-        if self.velocity_min is None:
-            self.velocity_min = np.ptp(self.BoundsArray,axis=1)/self.velocity_minimum_factor
+
+        if Velocity_clipping_or_rescale == 'Clip':
+            self.velocity_clipping_function = self.velocity_clipping
+            self.velocity_minimum_factor = Velocity_minimum_factor
+            self.velocity_min = Velocity_min
+            if self.velocity_min == None:
+                self.velocity_min = np.ptp(self.BoundsArray, axis=1) / self.velocity_minimum_factor
+    
+        elif Velocity_clipping_or_rescale == 'Rescale':
+
+            self.velocity_clipping_function = self.rescale_velocities
+            self.delta_max = Delta_max
+            self.delta_min = Delta_min
+            self.velocity_max = self.delta_max*np.linalg.norm(np.ptp(self.BoundsArray,axis=1))
+            self.velocity_min = self.delta_min*np.linalg.norm(np.ptp(self.BoundsArray,axis=1))
+            self.clip_lower_velocity = Clip_lower_velocity
+            self.clip_upper_velocity = Clip_upper_velocity
+
+        #Wether to reinitialise velocities from initial guess or from v_min array provided by user
+        self.reinitialise_velocities_from_initial_guess = Reinitialise_velocities_from_initial_guess
+
 
         if self.MH_fraction == 0:
             # The velocity rule is the PSO standard rule when no MH
@@ -238,18 +285,31 @@ class Swarm(object):
                         self.Points[i] = self.InitialGuess[i]
                     elif isinstance(self.InitialGuess[i], (dict)):
                         self.Points[i] = np.array([ self.InitialGuess[i][name] for name in self.Model.names ])
+                # If initial guess provided, velocities drawn from spread of initial distribution
+                # Spread in velocities according to the normal distribution specified by the covariance of initial positions
+                cov = np.cov(self.Points.T)
+                self.Velocities = np.random.multivariate_normal(np.zeros(self.Ndim), cov, size=self.NumParticles)*self.initial_guess_v_factor/np.sqrt(self.Ndim)
+
+                if self.reinitialise_velocities_from_initial_guess = True:
+                    if self.velocity_clipping_function == self.rescale_velocities:
+                        # set up velocity rescaling bounds from initial distribution if using the rescaling
+                        self.velocity_max = self.delta_max * np.linalg.norm(np.ptp(self.Points, axis=0))
+                        self.velocity_min = self.delta_min * np.linalg.norm(np.ptp(self.Points, axis=0))
+                    elif self.velocity_clipping_function == self.velocity_clipping:
+                        # set up minimum velocity bounds using initial distribution if using the rescaling
+                        self.velocity_min = np.ptp(self.Points, axis=0) / self.velocity_minimum_factor
+                else:
+                    # v_min already set in init method, above lines overwrite if set to initialise velocities from initial guess.
+                    pass
 
             # Initialise the particle function values
             self.Pool = Pool(self.Nthreads)
             self.Values = np.array( self.Pool.map(self.MyFunc, self.Points) )
 
-            # p.close()
-
             # Initialise each particle's best known position to initial position
             self.BestKnownPoints = self.Points.copy()
             self.BestKnownValues = self.Values.copy()
 
-
             # Update the swarm's best known position
             self.BestKnownSwarmPoint = self.BestKnownPoints[np.argmax(self.BestKnownValues)]
 
@@ -286,7 +346,8 @@ class Swarm(object):
 
     def PSO_VelocityRule(self):
         """
-        The standard PSO rule for updating the velocities
+        The standard PSO rule for updating the velocities with a jitter component added to reduce the chance of a particle
+        stuck in a local maxima
 
         RETURN:
         ------
@@ -299,14 +360,82 @@ class Swarm(object):
                                   ).reshape((self.NumParticles, self.Ndim))
         unclipped_velocities = (self.Omega * self.Velocities
                + self.PhiP * np.random.uniform(size=(self.NumParticles,self.Ndim)) * ( self.BestKnownPoints - self.Points )
-               + self.PhiG * np.random.uniform(size=(self.NumParticles,self.Ndim)) * ( best_known_swarm_point - self.Points))
-
-        # Clip velocities by the minimum velocity for each dimension to avoid pointless exploration
-        #   Need to compare absolute velocities (why we have to do the np.sign business)
-        clipped_velocities = np.sign(unclipped_velocities)*np.clip(np.abs(unclipped_velocities), a_min = self.velocity_min, a_max= None)
-
+               + self.PhiG * np.random.uniform(size=(self.NumParticles,self.Ndim)) * ( best_known_swarm_point - self.Points)
+               # line "jitters" each particle towards another random particle in the swarm to prevent being stuck in local maxima
+               + self.Jitter_weight * np.random.uniform(size=(self.NumParticles,self.Ndim)) * (self.Points[np.random.randint(self.NumParticles,size=self.NumParticles)]-
+                                                                                               self.Points))
+        clipped_velocities = self.velocity_clipping_function(unclipped_velocities)
         return (clipped_velocities)
 
+    def velocity_clipping(self,unclipped_velocities):
+        """
+        Clip the velocities only at the lower end, an alternative to velocity rescaling method
+
+        Use this method in problems where you REALLY dont want the velocity for 1 or more parameters to be too small as this
+            is possible with the resclaing method
+        INPUT:
+        ----------
+        unclipped_velocities: ndarray (#Num_particles,#Ndim)
+            PSO velocities unclipped
+
+        RETURN:
+        -------
+        clipped_velocities: ndarray (#Num_particles,#Ndim)
+            PSO velocities clipped
+        """
+        clipped_velocities = np.sign(unclipped_velocities)*np.clip(np.abs(unclipped_velocities), a_min = self.velocity_min, a_max= None)
+        return(clipped_velocities)
+
+    def rescale_velocities(self,unclipped_velocities):
+        """
+        Rescale velocities (on the upper  and lower end) by magnitude preserving particle directions.
+        (See https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=8280887)
+
+        INPUT:
+        ----------
+        unclipped_velocities: ndarray (#Num_particles,#Ndim)
+            PSO velocities unclipped
+
+        RETURN:
+        -------
+        clipped_velocities: ndarray (#Num_particles,#Ndim)
+            PSO velocities rescaled
+        """
+        # Calculate velocity magnitudes for each particle
+        velocity_magnitudes_uncliped = np.linalg.norm(np.abs(unclipped_velocities),axis=1)
+
+        # Clip manitudes by maximum mangitudes and minimum and detect particles where cliping was applied
+        # Seperate the lower and upper edge clipping
+        #    //TODO: probably can clean this up into one expression but not urgent
+        velocity_magnitudes_clipped_upper= np.clip(velocity_magnitudes_uncliped, a_min = None, a_max=self.velocity_max)
+        velocity_clipping_indices_upper = np.where(velocity_magnitudes_uncliped != velocity_magnitudes_clipped_upper)[0]
+
+        velocity_magnitudes_clipped_lower= np.clip(velocity_magnitudes_uncliped, a_min = self.velocity_min, a_max= None)
+        velocity_clipping_indices_lower = np.where(velocity_magnitudes_uncliped != velocity_magnitudes_clipped_lower)[0]
+
+        print('Number of velocity magnitudes rescaled (lower,upper): ',(velocity_clipping_indices_lower.size,velocity_clipping_indices_upper.size))
+        # Clipped velocities will be of the same shape as unclipped velocities
+        clipped_velocities = unclipped_velocities.copy()
+
+        # Check if any clipping occurs and reasign clipped velocities appropriately
+        if (velocity_clipping_indices_upper.size != 0) and self.clip_upper_velocity == True:
+            # Repeat magnitudes across all components of velocity for division
+            #      [[ v_particle_1_mag, v_particle_1_mag, .... v_particle_1_mag],
+            #        [v_particle_2_mag, v_particle_2_mag, .... v_particle_2_mag]...]
+            # As all dimensions divided by same dimension
+            # v magnitudes for each particle that needs to be clipped across all dimensions
+            magnitudes_reshaped_upper = np.repeat(velocity_magnitudes_uncliped[velocity_clipping_indices_upper][:,np.newaxis],repeats=self.Ndim, axis=1)
+
+            clipped_velocities[velocity_clipping_indices_upper] = unclipped_velocities[velocity_clipping_indices_upper]/magnitudes_reshaped_upper * self.velocity_max
+
+        # Same thing but for lower edge
+        if (velocity_clipping_indices_lower.size != 0) and self.clip_lower_velocity == True:
+
+            magnitudes_reshaped_lower = np.repeat(velocity_magnitudes_uncliped[velocity_clipping_indices_lower][:,np.newaxis],repeats=self.Ndim, axis=1)
+            clipped_velocities[velocity_clipping_indices_lower] = unclipped_velocities[velocity_clipping_indices_lower]/magnitudes_reshaped_lower * self.velocity_min
+
+
+        return(clipped_velocities)
     def Hybrid_VelocityRule(self):
         """
         The PSO rule for updating the velocities including a MH MCMC part to it
@@ -421,7 +550,6 @@ class Swarm(object):
         output_str += "Spread: {0}".format(np.ptp(self.Values))
         print(output_str)
 
-
     def SaveFinalResults(self):
         """
         Save the final results to file
@@ -440,6 +568,14 @@ class Swarm(object):
         file1.write(output_str)
         file1.close()
 
+        results_dictionary = {}
+        results_dictionary['Labels'] = self.Model.names
+        results_dictionary['Max Value'] = self.BestKnownSwarmValue
+        results_dictionary['Best point'] = self.BestKnownSwarmPoint
+        results_dictionary['Covariance matrix'] = np.cov(self.Points.T)
+        results_dictionary['Uncertainties'] = np.sqrt(np.diag(np.cov(self.Points.T)))
+        results_dictionary_file = os.path.join(self.Output, "Results.npz")
+        np.savez(results_dictionary_file, **results_dictionary)
 
 
     def ContinueCondition(self):
@@ -447,7 +583,7 @@ class Swarm(object):
         When continue condition ceases to be satisfied the evolution stops
         """
         spread = np.ptp(self.Values)
-        return ( spread>self.Tol  and  self.EvolutionCounter<self.MaxIter )
+        return (self.EvolutionCounter<self.MaxIter )
 
 
     def CreateEvolutionHistoryFile(self):
@@ -508,6 +644,19 @@ class Swarm(object):
         plt.savefig(outfile)
         plt.clf()
 
+        # Trajectory for each pair of params
+        for j, name_x in enumerate(self.Model.names):
+            for name_y in self.Model.names[j+1:]:
+                plt.figure()
+                for i in range(self.NumParticles):
+                    traj = np.array(swarm_points[i::self.NumParticles])
+                    plt.plot(traj[:,1+self.Model.names.index(name_x)], traj[:,1+self.Model.names.index(name_y)],
+                             '-', marker='o', markersize=3, color=palette[i], alpha=0.5)
+                plt.xlabel(name_x)
+                plt.ylabel(name_y)
+                outfile = os.path.join(self.Output, "EvolutionTrajectory_{0}_{1}.png".format(name_x, name_y))
+                plt.savefig(outfile)
+                plt.clf()
 
     def Run(self, segmenting=False):
         """
@@ -518,7 +667,6 @@ class Swarm(object):
             flag to indicate if we are transferring likelihoods
         """
         if segmenting==False: self.InitialiseSwarm()
-
         if self.SaveEvolution and self.EvolutionCounter==0:
             self.CreateEvolutionHistoryFile()
             self.SaveSwarmEvolution()
